// Image class implementation
#include "pch.h"

#include "Image.h"
#include "util.h"
#include "Buffer.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image.h"

namespace nf::client::render {
    Image::Image(ImageType type, const VkDevice& device, VideoMemoryAllocator& allocator, const VkCommandPool& commandPool, const VkQueue& queue, const std::string& imageData, VkExtent2D attachmentExtent)
        : GraphicsResource(device)
        , m_allocator(allocator)
        , m_handle()
        , m_allocation()
        , m_view()
    {
        VkFormat imageFormat = VK_FORMAT_UNDEFINED;

        switch (type) {
            case ImageType::Texture:
                imageFormat = VK_FORMAT_R8G8B8A8_SRGB;
                createTextureImage(imageFormat, commandPool, queue, imageData);
                break;

            case ImageType::DepthAttachment:
                imageFormat = VK_FORMAT_D32_SFLOAT;
                createImage(imageFormat, attachmentExtent, VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT);
                break;
        }

        createImageView(imageFormat);
    }

    const VkImageView& Image::getView() const {
        return m_view;
    }

    void Image::createTextureImage(VkFormat imageFormat, const VkCommandPool& commandPool, const VkQueue& queue, const std::string& imageData) {
        int imageWidth = 0, imageHeight = 0, numChannels = 0;
        stbi_uc* rawImageData = stbi_load_from_memory(reinterpret_cast<const stbi_uc*>(imageData.data()), imageData.size(), &imageWidth, &imageHeight, &numChannels, STBI_rgb_alpha);

        VkDeviceSize imageSize = static_cast<VkDeviceSize>(imageWidth) * imageHeight * 4;

        Buffer stagingBuffer(BufferType::Staging, m_device, m_allocator, nullptr, nullptr, rawImageData, imageSize);

        stbi_image_free(rawImageData);

        VkExtent2D imageExtent = { static_cast<uint32_t>(imageWidth), static_cast<uint32_t>(imageHeight) };
        createImage(imageFormat, imageExtent, VK_IMAGE_USAGE_TRANSFER_DST_BIT | VK_IMAGE_USAGE_SAMPLED_BIT);

        VkCommandBuffer commandBuffer = nullptr;
        VkCommandBufferAllocateInfo commandBufferAI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO };
        commandBufferAI.commandPool = commandPool;
        commandBufferAI.commandBufferCount = 1;
        vkAllocateCommandBuffers(m_device, &commandBufferAI, &commandBuffer);

        VkCommandBufferBeginInfo commandBufferBI = { VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO };
        commandBufferBI.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
        vkBeginCommandBuffer(commandBuffer, &commandBufferBI);

        // Image transition undefined -> transfer destination
        VkImageMemoryBarrier imageMemoryBarrier = { {VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER} };
        imageMemoryBarrier.image = m_handle;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
        imageMemoryBarrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
        imageMemoryBarrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
        imageMemoryBarrier.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        imageMemoryBarrier.subresourceRange.levelCount = 1;
        imageMemoryBarrier.subresourceRange.layerCount = 1;
        imageMemoryBarrier.srcAccessMask = 0;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT, VK_PIPELINE_STAGE_TRANSFER_BIT, 0, 0, nullptr, 0, nullptr, 1, &imageMemoryBarrier);

        vkEndCommandBuffer(commandBuffer);

        VkSubmitInfo submitInfo = { VK_STRUCTURE_TYPE_SUBMIT_INFO };
        submitInfo.commandBufferCount = 1;
        submitInfo.pCommandBuffers = &commandBuffer;
        vkQueueSubmit(queue, 1, &submitInfo, nullptr);
        vkQueueWaitIdle(queue);

        vkBeginCommandBuffer(commandBuffer, &commandBufferBI);

        VkBufferImageCopy bufferImageCopyRegion = {};
        bufferImageCopyRegion.imageSubresource.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        bufferImageCopyRegion.imageSubresource.layerCount = 1;
        bufferImageCopyRegion.imageExtent = { static_cast<uint32_t>(imageWidth), static_cast<uint32_t>(imageHeight), 1 };
        vkCmdCopyBufferToImage(commandBuffer, stagingBuffer.getHandle(), m_handle, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &bufferImageCopyRegion);

        vkEndCommandBuffer(commandBuffer);
        vkQueueSubmit(queue, 1, &submitInfo, nullptr);
        vkQueueWaitIdle(queue);

        vkBeginCommandBuffer(commandBuffer, &commandBufferBI);

        // Image transition transfer destination -> shader read only
        imageMemoryBarrier.oldLayout = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
        imageMemoryBarrier.newLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        imageMemoryBarrier.srcAccessMask = VK_ACCESS_TRANSFER_WRITE_BIT;
        imageMemoryBarrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
        vkCmdPipelineBarrier(commandBuffer, VK_PIPELINE_STAGE_TRANSFER_BIT, VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT, 0, 0, nullptr, 0, nullptr, 1, &imageMemoryBarrier);

        vkEndCommandBuffer(commandBuffer);
        vkQueueSubmit(queue, 1, &submitInfo, nullptr);
        vkQueueWaitIdle(queue);

        vkFreeCommandBuffers(m_device, commandPool, 1, &commandBuffer);
    }

    void Image::createImage(VkFormat format, VkExtent2D& size, VkImageUsageFlags usage) {
        VkImageCreateInfo imageCI = { VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO };
        imageCI.imageType = VK_IMAGE_TYPE_2D;
        imageCI.extent = { size.width, size.height, 1 };
        imageCI.mipLevels = 1;
        imageCI.arrayLayers = 1;
        imageCI.samples = VK_SAMPLE_COUNT_1_BIT;
        imageCI.format = format;
        imageCI.usage = usage;
        imageCI.tiling = VK_IMAGE_TILING_OPTIMAL;

        if (vkCreateImage(m_device, &imageCI, nullptr, &m_handle) != VK_SUCCESS)
            NFError("Could not create image.");

        m_allocator.allocateForImage(m_handle, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT, m_allocation);
    }

    void Image::createImageView(VkFormat format) {
        VkImageAspectFlags aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        if (format == VK_FORMAT_D32_SFLOAT)
            aspectMask = VK_IMAGE_ASPECT_DEPTH_BIT;

        VkImageViewCreateInfo imageViewCI = { VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO };
        imageViewCI.image = m_handle;
        imageViewCI.viewType = VK_IMAGE_VIEW_TYPE_2D;
        imageViewCI.format = format;
        imageViewCI.subresourceRange.aspectMask = aspectMask;
        imageViewCI.subresourceRange.levelCount = 1;
        imageViewCI.subresourceRange.layerCount = 1;

        if (vkCreateImageView(m_device, &imageViewCI, nullptr, &m_view) != VK_SUCCESS)
            NFError("Could not create image view.");
    }

    Image::~Image() {
        vkDestroyImageView(m_device, m_view, nullptr);
        m_allocator.deallocate(m_allocation);
        vkDestroyImage(m_device, m_handle, nullptr);
    }
}
